#!/bin/bash


# Starts Envoy and configures traffic interception on GCE VM to access
# GCP Traffic Director.
#
# Usage steps:
#   - Configure environment in "sidecar.env".
#   - Run this script as root: sudo /bin/bash -x </path/to/this/script>
#
# Notes:
#   - Envoy will be started in a Docker container. Only one Envoy container
#     created by this script can exist at any given time. Running ./run.sh stop
#     should stop and remove any existing
#     container.
#   - This script must be run as root, as required by iptables and Docker (for a
#     typical install).


set -o errexit
set -o nounset
set -o pipefail

if [[ "${EUID}" -ne 0 ]]; then
  echo "$0 must be run with root privileges. Try sudo $0" >&2
  exit 1
fi

# Location of compulsory files.
readonly PACKAGE_DIRECTORY=$(dirname "$0")
readonly CONFIG_FILE="${PACKAGE_DIRECTORY}/sidecar.env"
readonly IPTABLES_SCRIPT="${PACKAGE_DIRECTORY}/iptables.sh"
readonly BOOTSTRAP_TEMPLATE="${PACKAGE_DIRECTORY}/bootstrap_template.yaml"
readonly BOOTSTRAP_INSTANCE="${PACKAGE_DIRECTORY}/bootstrap_instance.yaml"
readonly DOCKERFILE_TEMPLATE="${PACKAGE_DIRECTORY}/Dockerfile_template"
# To prevent redirection to the metadata server to enable ssh via GCP UI.
readonly METADATA_SERVER_CIDR="169.254.169.254/32"
# The name this script will use for its local Envoy image.
readonly TD_ENVOY_IMAGE="td_envoy_image"
# The name this script will use to identify the Envoy container.
ENVOY_CONTAINER_NAME='td_envoy'

function initialize() {
  # Read configuration.
  source "${CONFIG_FILE}"

  # Set optional variables if they are not defined previously.
  TRACING_ENABLED="${TRACING_ENABLED:-false}"
  ACCESSLOG_PATH="${ACCESSLOG_PATH:-}"
  ENVOY_ADMIN_PORT="${ENVOY_ADMIN_PORT:-15000}"
  BACKEND_INBOUND_PORTS="${BACKEND_INBOUND_PORTS:-}"
  INTERCEPT_INBOUND_PORTS="${INTERCEPT_INBOUND_PORTS:-}"

  if ! docker --help &> /dev/null; then
    echo 'Docker is required to run Envoy in a container. Please see https://docs.docker.com/engine/install/ for installation steps.' >&2
    exit 1
  fi

  if [[ -z "${ENVOY_USER}" ]]; then
    echo "ENVOY_USER must be set in ${CONFIG_FILE}." >&2
    exit 1
  fi
  ENVOY_UID=$(id -u "${ENVOY_USER}")

  if [[ -n "${BACKEND_INBOUND_PORTS}" ]] && [[ "${BACKEND_INBOUND_PORTS}" != "${INTERCEPT_INBOUND_PORTS}" ]]; then
    echo -n "WARNING: INTERCEPT_INBOUND_PORTS is not set while BACKEND_INBOUND_PORTS is set. iptables will not be configured to redirect inbound traffic to Envoy." >&2
  fi

  if [[ -z "${VPC_NETWORK_NAME}" ]]; then
    echo "VPC_NETWORK_NAME must be set in ${CONFIG_FILE}." >&2
    exit 1
  fi

  if [[ -z "${GCP_PROJECT_NUMBER}" ]]; then
    echo "GCP_PROJECT_NUMBER must be set in ${CONFIG_FILE}." >&2
    exit 1
  fi

  chmod +x "${IPTABLES_SCRIPT}"
}

function enable_interception() {
  # Run iptables.sh.
  echo "Enabling traffic interception for the '${SERVICE_CIDR}' destination."
  TRAFFIC_DIRECTOR_GCE_VM_DEPLOYMENT_OVERRIDE='true'
  DISABLE_REDIRECTION_ON_LOCAL_LOOPBACK='true'
  export TRAFFIC_DIRECTOR_GCE_VM_DEPLOYMENT_OVERRIDE
  export DISABLE_REDIRECTION_ON_LOCAL_LOOPBACK
  declare -a IPTABLES_ARGS
  IPTABLES_ARGS=( "-x" "${METADATA_SERVER_CIDR}" "-i" "${SERVICE_CIDR}" \
    "-p" "${ENVOY_PORT}" )
  if [[ -n "${EXCLUDE_ENVOY_USER_FROM_INTERCEPT}" ]]; then
    IPTABLES_ARGS+=( "-u" "${ENVOY_UID}" )
  fi
  if [[ -n "${INTERCEPT_INBOUND_PORTS}" ]]; then
    IPTABLES_ARGS+=( "-m" "REDIRECT" "-b" "${INTERCEPT_INBOUND_PORTS}" )
  fi
  "${IPTABLES_SCRIPT}" "${IPTABLES_ARGS[@]}" > /dev/null 2>&1
}

function disable_interception() {
  echo "Disabling traffic interception for the '${SERVICE_CIDR}' destination."
  "${IPTABLES_SCRIPT}" clean > /dev/null 2>&1
}

function get_zone() {
  # Query the GCP instance metadata server to find the zone of the current VM.
  # Delete any iptables rules created by previous executions.
  disable_interception > /dev/null
  envoy_zone=''
  query_url='http://metadata.google.internal/computeMetadata/v1/instance/zone'
  query_header='Metadata-Flavor: Google'
  query_result=$(curl "${query_url}" -H "${query_header}" -sS || true)
  if [[ -z "${query_result}" ]]; then
    echo 'Failed to get zone from GCP metadata server.' >&2
  else
    # The response is in the format: "projects/[PROJECT_NUMBER]/zones/[ZONE]"
    envoy_zone=$(echo "${query_result}" | cut -d'/' -f4 || true)
    if [[ -z "${envoy_zone}" ]]; then
      echo 'Failed to parse the result from GCP metadata server.' >&2
    fi
  fi
}


function prepare_envoy() {
  # Get Envoy GCP zone.
  get_zone
  # Generate Envoy bootstrap.
  envoy_node_id="$(uuidgen)~$(hostname -i)"
  cat "${BOOTSTRAP_TEMPLATE}" \
      | sed -e "s|ENVOY_NODE_ID|${envoy_node_id}|g" \
      | sed -e "s|ENVOY_ZONE|${envoy_zone}|g" \
      | sed -e "s|VPC_NETWORK_NAME|${VPC_NETWORK_NAME}|g" \
      | sed -e "s|CONFIG_PROJECT_NUMBER|${GCP_PROJECT_NUMBER}|g" \
      | sed -e "s|ENVOY_PORT|${ENVOY_PORT}|g" \
      | sed -e "s|ENVOY_ADMIN_PORT|${ENVOY_ADMIN_PORT}|g" \
      | sed -e "s|XDS_SERVER_CERT|${XDS_SERVER_CERT}|g" \
      | sed -e "s|TRACING_ENABLED|${TRACING_ENABLED}|g" \
      | sed -e "s|ACCESSLOG_PATH|${ACCESSLOG_PATH}|g" \
      | sed -e "s|BACKEND_INBOUND_PORTS|${BACKEND_INBOUND_PORTS}|g" \
      > "${BOOTSTRAP_INSTANCE}"

  # Build an image using the new bootstrap which mirrors the Envoy user so that
  # iptables may ignore it.
  cat "${DOCKERFILE_TEMPLATE}" \
      | sed -e "s|BOOTSTRAP_INSTANCE|${BOOTSTRAP_INSTANCE}|g" \
      | sed -e "s|ENVOY_IMAGE|${ENVOY_IMAGE}|g" \
      > "Dockerfile"
  docker build . -t "${TD_ENVOY_IMAGE}"

  # Grant permissions to ${ENVOY_USER}.
  mkdir -p "${LOG_DIR}"
  chown -R "${ENVOY_USER}": "${LOG_DIR}"
}

function start_envoy() {
  echo "Starting Envoy container."
  docker run -d \
    --name="${ENVOY_CONTAINER_NAME}" \
    --network="host" \
    --volume="${LOG_DIR}:/mnt/envoy_logs/" \
    ${ACCESSLOG_PATH:+--volume="$(dirname "${ACCESSLOG_PATH}"):$(dirname "${ACCESSLOG_PATH}")"} \
    -e "ENVOY_UID=${ENVOY_UID}" \
    "${TD_ENVOY_IMAGE}" \
    envoy \
      --config-path /etc/envoy/envoy.yaml \
      --log-level "${LOG_LEVEL}" \
      --log-path "/mnt/envoy_logs/envoy.log" \
      --allow-unknown-fields
}

function stop_envoy() {
  echo "Stopping Envoy container."
  docker stop "${ENVOY_CONTAINER_NAME}"
  echo "Removing Envoy container."
  docker rm "${ENVOY_CONTAINER_NAME}"
  echo "Removing local Envoy image."
  docker image rm "${TD_ENVOY_IMAGE}"
}

function check_status() {
  ENVOY_IS_RUNNING=0

  container_exist=$(docker ps -q -a -f name=${ENVOY_CONTAINER_NAME})
  if [[ -z "${container_exist}" ]]; then
    echo "NOT ENABLED: Envoy container does not exist."
  else
    # If the container exists, this will emit true/false based on running status.
    envoy_status=$(docker container inspect -f "{{.State.Running}}" ${ENVOY_CONTAINER_NAME} || printf 'missing')
    case "${envoy_status}" in
      'true')
        echo "OK: Envoy container seems to be running."
        ENVOY_IS_RUNNING=1
        ;;
      'false')
        echo "NOT ENABLED: Envoy container is stopped."
        ;;
      *"missing") # The * here accounts for the output of inspect.
        echo "NOT ENABLED: Envoy container does not exist."
        ;;
      *)
        echo "NOT ENABLED: Unknown Envoy container status: ${envoy_status}"
        ;;
    esac
  fi

  if [[ $(iptables -t nat -S | grep -c ISTIO) -gt 0 ]]; then
    echo "OK: Traffic interception seems to be enabled."
  else
    echo "NOT ENABLED: Traffic interception does not seem to be enabled."
  fi
}

initialize

case ${1-usage} in
  'start')
    check_status > /dev/null
    if [[ "$ENVOY_IS_RUNNING" -eq 1 ]]; then
      echo "Envoy is already running. Will not start another one." >&2
      echo "You can stop it by running 'sudo $0 stop'." >&2
    else
      # If envoy container exists, delete it before starting a new one.
      container_exist=$(docker ps -q -a -f name=${ENVOY_CONTAINER_NAME})
      if [[ "${container_exist}" ]]; then
        stop_envoy
      fi
      prepare_envoy
      start_envoy
    fi
    enable_interception
    check_status
    ;;
  'stop')
    disable_interception
    stop_envoy
    ;;
  'status')
    check_status
    ;;
   *)
    echo "Usage: $0 <start|stop|status>" >&2
    ;;
esac
